#include "config.h"

#include "types.h"
#include "io.h"
#include "display.h"

#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <time.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

display d;

int displayInit(void)
{
	d.start_time = 0;
	d.total_time = 0;
	d.current_time = 0;
	d.elapsed_time = 0;
	d.percent_complete = 0.0;
	d.display_interval = 1;
	d.overtime_flag = 0;
	d.k = 1024;
	d.twiddle = '-';
	d.screen_width = 79;
	d.display_twiddle = 1;
	d.display_datacount = 1;
	d.display_throughput = 1;
	d.display_time = 1;
	d.display_percent = 1;
	d.display_bar = 1;
	d.display_summary = 1;
	return(0);
}

int displayBegin(void)
{
	d.start_time = time(0);
	d.current_time = time(0);

	return(0);
}

#define sec_per_hour 3600
#define sec_per_minute 60

int calculateByteDisplay(uint64 *wptr,
	char **wXptr,
	float *wPptr
	)
{
	uint64 num = 0;
	uint64 div = 1;

	*wptr = io.total_write;
	*wXptr = "b ";
	*wPptr = 0.0;

	num = io.total_write;
	if (num >= d.k)         { *wXptr = "Kb"; div = d.k; }
	if (num >= (d.k * d.k)) { *wXptr = "Mb"; num /= d.k; }
	if (num >= (d.k * d.k)) { *wXptr = "Gb"; num /= d.k; }
	if (num >= (d.k * d.k)) { *wXptr = "Tb"; num /= d.k; }
	if (num >= (d.k * d.k)) { *wXptr = "Pb"; num /= d.k; }
	if (num >= (d.k * d.k)) { *wXptr = "Eb"; num /= d.k; }
	*wPptr = ((float)num / div);
	return(0);
}

int calculateBpsDisplay(uint64 *bpsptr,
	char **bpsXptr,
	float *bpsPptr
	)
{
	uint64 num = 0;
	uint64 div = 1;

	if (d.elapsed_time > 0) {
		*bpsptr = io.total_write / d.elapsed_time;
		num = io.total_write / d.elapsed_time;
	}
	else {
		*bpsptr = 0;
		num = 0;
	}
	*bpsXptr = "bps ";
	*bpsPptr = 0.0;

	if (num >= d.k)         { *bpsXptr = "Kbps"; div = d.k; }
	if (num >= (d.k * d.k)) { *bpsXptr = "Mbps"; num /= d.k; }
	if (num >= (d.k * d.k)) { *bpsXptr = "Gbps"; num /= d.k; }
	if (num >= (d.k * d.k)) { *bpsXptr = "Tbps"; num /= d.k; }
	*bpsPptr = ((float)num / div);
	return(0);
}

int calculateTimeDisplay(uint64 *sptr, uint64 *mptr, uint64 *hptr)
{
	if (*sptr > sec_per_hour) {
		*hptr = *sptr / sec_per_hour;
		*sptr -= *hptr * sec_per_hour;
	}
	if (*sptr > sec_per_minute) {
		*mptr = *sptr / sec_per_minute;
		*sptr -= *mptr * sec_per_minute;
	}
	return(0);
}

int calculatePercentComplete(void)
{
	if (io.total_size_known && (io.total_size > 0)) {
		d.percent_complete = (float)(io.total_write * 100.0 / io.total_size);
		if (d.percent_complete < 0) {
			d.percent_complete = 9999.9;
		}
	}
	return(0);
}

void displayTwiddle(void)
{
	static int t = 0;

	if (!d.display_twiddle) return;
	if (d.elapsed_time == 0) return;
	t += io.total_write / d.elapsed_time;
	if (t < 10) return;
	t = 0;
	switch (d.twiddle) {
		case '\\': d.twiddle = '|';  break;
		case '|':  d.twiddle = '/';  break;
		case '/':  d.twiddle = '-';  break;
		case '-':  d.twiddle = '\\'; break;
	}
	fprintf(stderr, "%c\r", d.twiddle);
}

int displayPrint(void)
{
	uint64 eta = 0;
	uint64 bps = 0;
	char *bpsX = "bps";
	float bpsP = 0.0;
	uint64 hours = 0;
	uint64 minutes = 0;
	uint64 seconds = 0;
	uint64 w = 0;
	char *wX = "";
	float wP = 0.0;
	char *time_title = "eta";
	int screen_used = 0;
	int this_width = 0;

	d.current_time = time(0);
	d.elapsed_time = d.current_time - d.start_time;

	if (calculateBpsDisplay(&bps, &bpsX, &bpsP) != 0) {
		return(1);
	}

	if (calculatePercentComplete() != 0) {
		return(1);
	}

	if (io.total_size_known == 1) {
		if (io.total_write > 0) {
			if (io.total_size >= io.total_write) {
				if (d.percent_complete > 0.0) {
					eta = 
						(uint64)(
							100 * d.elapsed_time / d.percent_complete
						)
						- d.elapsed_time;
				}
				else {
					eta = (uint64)(-1);
				}
			}
			else {
				if (!d.overtime_flag) {
					d.overtime_flag = 1;
					d.total_time = d.elapsed_time;
				}
				eta = d.elapsed_time - d.total_time;
				time_title = "ovr";
			}
		}
		else {
			eta = 0;
		}
		seconds = eta;
	}
	else
	{
		seconds = d.elapsed_time;
		time_title = "elapsed";
	}
	if (calculateTimeDisplay(&seconds, &minutes, &hours) != 0) {
		return(1);
	}

	if (calculateByteDisplay(&w, &wX, &wP) != 0) {
		return(1);
	}

	/*
	 * Display twiddle
	 */
	this_width = 1;
	if ((d.display_twiddle) && (screen_used+this_width < d.screen_width)) {
		fprintf(stderr, "%c", d.twiddle);
		screen_used += this_width;
	}

	/*
	 * Display data count
	 */
	this_width = 8;
	if (screen_used > 0)
		this_width++;
	if ((d.display_datacount) && (screen_used+this_width < d.screen_width)) {
		if (screen_used > 0)
			fprintf(stderr, " ");
		if (wP > 9999.9) {
			fprintf(stderr, "+999.9%2.2s", wX);
		}
		else {
			fprintf(stderr, "%6.1f%2.2s", wP, wX);
		}
		screen_used += this_width;
	}
		
	/*
	 * Display throughput
	 */
	this_width = 13;
	if (!d.display_datacount)
		this_width -= 3;
	if (screen_used > 0)
		this_width++;
	if ((d.display_throughput) && (screen_used+this_width < d.screen_width)) {
		if (screen_used > 0)
			fprintf(stderr, " ");
		if (d.display_datacount)
			fprintf(stderr, "at ");
		fprintf(stderr, "%6.1f%4.4s", bpsP, bpsX);
		screen_used += this_width;
	}

	/*
	 * Display time
	 */
	this_width = 11+strlen(time_title);
	if (screen_used > 0)
		this_width += 2;
	if ((d.display_time) && (screen_used+this_width < d.screen_width)) {
		if (screen_used > 0)
			fprintf(stderr, "  ");
		fprintf(stderr, "%s: ", time_title);
		if (hours > 99) {
			fprintf(stderr, "+99:99:99");
		}
		else {
			fprintf(stderr, "%3llu:%2.2llu:%2.2llu", hours, minutes, seconds);
		}
		screen_used += this_width;
	}

	/*
	 * Display percent
	 */
	this_width = 5;
	if (screen_used > 0)
		this_width++;
	if ((d.display_percent) && (io.total_size_known)
		&& (screen_used+this_width < d.screen_width))
	{
		if (screen_used > 0)
			fprintf(stderr, " ");
		if (d.percent_complete > 999) {
			fprintf(stderr, "+999%%");
		}
		else {
			fprintf(stderr, "%4d%%", (int)d.percent_complete);
		}
		screen_used += this_width;
	}

	/*
	 * Display progress bar
	 */
	this_width = 5;
	if (screen_used > 0)
		this_width++;
	if ((d.display_bar) && (io.total_size_known)
		&& (screen_used+this_width < d.screen_width))
	{
		int c;
		int line_length;
		int completed_length = 0;

		if (screen_used > 0) {
			fprintf(stderr, " ");
			screen_used++;
		}
		this_width = d.screen_width - screen_used + 1;
		line_length = this_width - 3;
		completed_length = line_length * d.percent_complete / 100;
		fprintf(stderr, "[");
		for (c = 0; c < line_length; c++) {
			if (c <= completed_length) {
				fprintf(stderr, "=");
			}
			else {
				fprintf(stderr, " ");
			}
		}
		fprintf(stderr, "]");
	}

	fprintf(stderr, "\r");
	return(0);
}

int displayUpdate(void)
{
	displayTwiddle();
	if (time(0) < d.current_time + d.display_interval) {
		fprintf(stderr, "\r");
		return(0);
	}
	if (displayPrint() != 0) return(1);
	return(0);
}

int displayEnd(void)
{
	uint64 bps = 0;
	char *bpsX = "bps";
	float bpsP = 0.0;
	uint64 w = 0;
	char *wX = "";
	float wP = 0.0;
	uint64 hours = 0;
	uint64 minutes = 0;
	uint64 seconds = 0;

	displayPrint();
	d.elapsed_time = d.current_time - d.start_time;

	if (calculatePercentComplete() != 0) {
		return(1);
	}

	if (calculateByteDisplay(&w, &wX, &wP) != 0) {
		return(1);
	}

	if (calculateBpsDisplay(&bps, &bpsX, &bpsP) != 0) {
		return(1);
	}

	seconds = d.elapsed_time;
	if (calculateTimeDisplay(&seconds, &minutes, &hours) != 0) {
		return(1);
	}

	fprintf(stderr, "\n");
	if (d.display_summary) {
		fprintf(stderr, "Bytes Copied: %llu (%.1f%s)", w, wP, wX);
		if (io.total_size_known) {
			fprintf(stderr, " (%d%% of expected input)", (int)d.percent_complete);
		}
		fprintf(stderr, "\n");

		fprintf(stderr, "Time: ");
		if (hours > 0) {
			fprintf(stderr, "%3llu:%2.2llu:%2.2llu", hours, minutes, seconds);
		}
		else if (minutes > 0) {
			fprintf(stderr, "%2.2llu:%2.2llu", minutes, seconds);
		}
		else {
			fprintf(stderr, "%2llu seconds", seconds);
		}
		fprintf(stderr, "\n");

		if ((hours != 0) || (minutes != 0) || (seconds != 0)) {
			fprintf(stderr, "Throughput: %llubps (%.1f%s)\n\n", bps, bpsP, bpsX);
		}
		else {
			fprintf(stderr, "Throughput: (infinite)\n\n");
		}
	}

	d.start_time = 0;
	
	return(0);
}
